import numpy as np
from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.utils.logger_factory import LoggerFactory


class BaseScaler(IProcessBlock):
    """
    Scales point cloud by centering it by its mean and dividing by the radius of the cloud
    """
    mean: np.ndarray = np.zeros((1, 3))
    scale: float = 1.0

    def __init__(self):
        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )


from typing import Tuple

import numpy as np
import open3d as o3d

from or_pcd.Optimizer.iOptimizer import IOptimizer
from or_pcd.utils.constants import (
    __MAXIMUM_CORRESPONDENCE_DISTANCE__,
    __MAX_ITERATIONS__,
)
from or_pcd.utils.logger_factory import LoggerFactory


class GeneralizedICP(IOptimizer):
    def __init__(
        self,
        max_correspondence_distance: float = __MAXIMUM_CORRESPONDENCE_DISTANCE__,
        max_iterations: int = __MAX_ITERATIONS__,
    ):
        """
        Refer to this link: https://www.robots.ox.ac.uk/~avsegal/resources/papers/Generalized_ICP.pdf for parameters explanation
        :param max_correspondence_distance:
        :param max_iterations:
        """

        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )
        if max_correspondence_distance <= 0:
            msg = f"max correspondence distance cannot be 0 or less. Provided: {max_correspondence_distance}"
            self._LOG.warning(msg)
            self._max_correspondence_distance = __MAXIMUM_CORRESPONDENCE_DISTANCE__
        else:
            self._max_correspondence_distance = max_correspondence_distance

        if max_iterations <= 0:
            msg = f"max iterations cannot be 0 or less. Provided: {max_iterations}"
            self._LOG.warning(msg)
            self._max_iterations = __MAX_ITERATIONS__
        else:
            self._max_iterations = max_iterations

        self._LOG.debug(
            f"Initialized {self.__class__.__name__} with parameters: {self}"
        )

    def optimize(
        self, source: np.ndarray, target: np.ndarray, **kwargs
    ) -> Tuple[np.ndarray, float]:
        icp_type = (
            o3d.pipelines.registration.TransformationEstimationForGeneralizedICP()
        )

        source_point_cloud = o3d.geometry.PointCloud()
        target_point_cloud = o3d.geometry.PointCloud()
        source_point_cloud.points = o3d.utility.Vector3dVector(source)
        target_point_cloud.points = o3d.utility.Vector3dVector(target)

        optimization_result: o3d.pipelines.registration.RegistrationResult = (
            o3d.pipelines.registration.registration_generalized_icp(
                source_point_cloud,
                target_point_cloud,
                self._max_correspondence_distance,
                np.eye(4),
                icp_type,
                o3d.pipelines.registration.ICPConvergenceCriteria(
                    max_iteration=self._max_iterations
                ),
            )
        )

        roto_translation = np.copy(optimization_result.transformation)
        transposed_rotation = roto_translation[:3, :3].T
        roto_translation[:3, :3] = transposed_rotation

        # If the rmse is 0 raise an error as parameters are not correct
        if optimization_result.inlier_rmse == 0:
            msg = """Optimization failed with loss = 0. Parameters are not well set. 
        Probably due to maximum_correspondence_distance set too low."""
            self._LOG.error(msg)
            raise ValueError(msg)

        return roto_translation, optimization_result.inlier_rmse

    def __repr__(self):
        return f"""{self.__class__.__name__}
            (max_correspondence_distance={self._max_correspondence_distance},
            max_iterations={self._max_iterations})
        """



import numpy as np
from scipy.spatial.distance import cdist

from or_pcd.Preprocessor.Scalers.BaseScaler import BaseScaler
from or_pcd.utils.logger_factory import LoggerFactory


class RadiusScaler(BaseScaler):
    """
    Scales point cloud by centering it by its mean and dividing by the radius of the cloud
    """

    def __init__(self):
        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )

    def process(self, cloud: np.ndarray) -> np.ndarray:
        """
        IprocessBlock implementation that saves mean and scale of the cloud
        :param cloud: point-cloud to process
        :return: processed cloud
        """
        self._LOG.debug("Normalizing clouds")
        center = np.mean(cloud, axis=0, keepdims=True)
        radius = np.max(cdist(center, cloud))
        self.mean = center
        self.scale = radius

        return (cloud - center) / radius



import copy
import multiprocessing
import time
from typing import List
from typing import Tuple

import numpy as np
import open3d as o3d

from or_pcd.Optimizer.iOptimizer import IOptimizer
from or_pcd.Preprocessor.preprocessor import Preprocessor
from or_pcd.Preprocessor.Scalers.BaseScaler import BaseScaler
from or_pcd.Visualizer.Visualizer import (
    visualize_point_clouds,
    draw_registration_result,
)
from or_pcd.utils.constants import (
    __ALIGNER_DEG__,
    __ALIGNER_MU__,
    __ALIGNER_STD__,
    __ALIGNER_MAX_ITER__,
    __ALIGNER_DELTA__,
    __ALIGNER_EPS__,
    __MULTISTART_ATTEMPTS__,
    __REFINER_MAX_ITER__,
    __REFINER_DISTANCE_THRESHOLD__,
)
from or_pcd.utils.create_cloud import create_cloud
from or_pcd.utils.logger_factory import LoggerFactory


class Aligner:
    transfromation: np.ndarray = np.eye(4)
    scale_factors: np.ndarray = np.ones((1, 3))

    def __init__(
        self,
        source_preprocessor: Preprocessor,
        target_preprocessor: Preprocessor,
        optimizer: IOptimizer,
        attempts: int = __MULTISTART_ATTEMPTS__,
        deg: float = __ALIGNER_DEG__,
        mu: float = __ALIGNER_MU__,
        std: float = __ALIGNER_STD__,
        delta: float = __ALIGNER_DELTA__,
        max_iter: int = __ALIGNER_MAX_ITER__,
        eps: float = __ALIGNER_EPS__,
        visualize_intermediate_steps: bool = False,
    ):
        """
        Wrapper function that takes the inner block optimizer, the preprocessors and runs a multi-start optimization on
        the source and target cloud.
        :param source_preprocessor: Source cloud preprocessor
        :param target_preprocessor: Target cloud preprocessor
        :param optimizer: iOptimizer inherited class to run a single optimization
        :param attempts: number of multi-start attempts
        :param deg: initial rotation matrix angle
        :param mu: initial translation vector mean
        :param std: initial translation vector standard deviation
        :param delta: initial compass step
        :param max_iter: maximum number of iterations for the compass search
        :param eps: stopping criteria on delta for the compass search
        """

        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=True
        )

        if attempts <= 0:
            msg = f"attempts cannot be 0 or less. Provided: {attempts}"
            self._LOG.warning(msg)
            self._attempts = __MULTISTART_ATTEMPTS__
        else:
            self._attempts = attempts

        if deg <= 0:
            msg = f"deg cannot be 0 or less. Provided: {deg}"
            self._LOG.warning(msg)
            self._deg = __ALIGNER_DEG__
        else:
            self._deg = deg

        if mu < 0:
            msg = f"mu cannot be 0 or less. Provided: {mu}"
            self._LOG.warning(msg)
            self._mu = __ALIGNER_MU__
        else:
            self._mu = mu

        if std <= 0:
            msg = f"std cannot be 0 or less. Provided: {std}"
            self._LOG.warning(msg)
            self._std = __ALIGNER_STD__
        else:
            self._std = std

        if delta <= 0:
            msg = f"delta cannot be 0 or less. Provided: {delta}"
            self._LOG.warning(msg)
            self._delta = __ALIGNER_DELTA__
        else:
            self._delta = delta

        if max_iter <= 0:
            msg = f"max_iter cannot be 0 or less. Provided: {max_iter}"
            self._LOG.warning(msg)
            self._max_iter = __ALIGNER_MAX_ITER__
        else:
            self._max_iter = max_iter
        print("EPS", eps)
        if eps <= 0:
            msg = f"eps cannot be 0 or less. Provided: {eps}"
            self._LOG.warning(msg)
            self._eps = __ALIGNER_EPS__
        else:
            self._eps = eps

        self._visualize_intermediate_steps = visualize_intermediate_steps
        self._source_preprocessor = source_preprocessor
        self._target_preprocessor = target_preprocessor
        self._optimizer = optimizer

        self._LOG.debug(f"Initialized {self.__class__.__name__} with {self}")

    def initialize_rotation(self) -> Tuple[np.ndarray, np.ndarray]:
        """Generate a random transformation matrix for a single multi start registration run."""

        # Generate random angles
        theta_1 = np.random.uniform(low=-self._deg, high=self._deg)
        theta_2 = np.random.uniform(low=-self._deg, high=self._deg)
        theta_3 = np.random.uniform(low=-self._deg, high=self._deg)

        # Create the 3 basic rotation matrices
        r_1 = np.array(
            [
                [1, 0, 0],
                [0, np.cos(theta_1), -np.sin(theta_1)],
                [0, np.sin(theta_1), np.cos(theta_1)],
            ]
        )
        r_2 = np.array(
            [
                [np.cos(theta_2), 0, np.sin(theta_2)],
                [0, 1, 0],
                [-np.sin(theta_2), 0, np.cos(theta_2)],
            ]
        )
        r_3 = np.array(
            [
                [np.cos(theta_3), -np.sin(theta_3), 0],
                [np.sin(theta_3), np.cos(theta_3), 0],
                [0, 0, 1],
            ]
        )

        # Aggregate them to form the final rotation matrix
        rotation_matrix = np.dot(r_1, np.dot(r_2, r_3))

        # Generate random translation vector
        translation = self._mu + np.random.randn(3) * self._std

        return rotation_matrix, translation

    def multistart_registration(
        self, source: np.ndarray, target: np.ndarray
    ) -> Tuple[np.ndarray, float]:
        """Perform multi-start registration on the source and target point clouds.
        :param source: Source point cloud
        :param target: Target point cloud

        :return: (best_transformation, metric): (best transformation matrix, best metric)
        """

        # Initial metric and transformation
        metric = np.inf
        best_transformation = np.eye(4)

        for n in range(self._attempts):  # tqdm
            source_copy = copy.deepcopy(source)

            # Generate random rotation and translation matrices
            initial_rotation, initial_translation = self.initialize_rotation()
            source_initialized = (
                np.dot(source_copy, initial_rotation) + initial_translation
            )

            # Perform registration
            current_transform, current_metric = self._optimizer.optimize(
                source_initialized, target
            )

            if current_metric < metric:
                # Update metric
                metric = current_metric
                # Define Transformation
                T = np.eye(4)
                T[:3, :3] = np.dot(initial_rotation, current_transform[:3, :3])
                T[:3, 3] = (
                    np.dot(initial_translation, current_transform[:3, :3]).ravel()
                    + current_transform[:3, 3]
                )
                best_transformation = T

        return best_transformation, metric  # best_transformation, metric

    def compass_step(
        self,
        source: np.ndarray,
        target: np.ndarray,
        scale_factors: np.ndarray,
        delta: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray, float]:
        """Perform a single compass step to find the optimal scaling factor for the target cloud.
        :param source: Source point cloud
        :param target: Target point cloud
        :param scale_factors: Current scaling factor
        :param delta: Step size for the scaling factor

        :return: (new_scale_factors, new_rotation, metric): (new scaling factor, obtsined new_rotation ans new metric)
        """
        new_scale_factors = scale_factors + delta
        target_scaled = target * new_scale_factors
        current_rotation, current_metric = self.multistart_registration(
            source, target_scaled
        )
        return new_scale_factors, current_rotation, current_metric

    def align(
        self,
        source: np.ndarray,
        target: np.ndarray,
        refine_registration: bool = True,
        icp_type: str = "PointToPlane",
    ) -> Tuple[np.ndarray, float, np.ndarray, List[float]]:
        """Align the source and target point clouds. using a two block optimization approach.
        :param source: Source point cloud
        :param target: Target point cloud
        :param refine_registration: Flag to enable/disable the refinement step that uses ICP
        :param icp_type: PointToPoint or PointToPlane ICP

        :return: (optimal_transformation, optimal_metric, optimal_scale_factors, errors): (Optimal transformation matrix, RMSE of the inliers, optimal scaling factor, list of errors)
        """
        source = self._source_preprocessor.preprocess(source)
        target = self._target_preprocessor.preprocess(target)
        iteration = 0

        if self._visualize_intermediate_steps:
            visualize_point_clouds([source, target], [(0, 0, 1), (1, 0, 0)])

        optimal_scale_factors = np.ones((1, 3))

        start = time.time()
        optimal_transformation, optimal_metric = self.multistart_registration(
            source, target
        )
        self._LOG.info(f"Multi-start registration time: {time.time() - start}")

        # INITIALIZE ERRORS LIST
        errors = [optimal_metric]
        self._LOG.info(f"Initial RMSE: {optimal_metric}")
        directions = np.eye(3)

        while self._delta >= self._eps and iteration <= self._max_iter:
            # UPDATE ITERATION COUNTER
            iteration += 1
            self._LOG.info(f"Iteration number: {iteration} Current step: {self._delta}")
            for axis in range(3):
                scale_plus = self._delta * directions[:, axis]
                new_scale_factors, new_rotation, new_metric = self.compass_step(
                    source, target, optimal_scale_factors, scale_plus
                )

                if new_metric <= optimal_metric:
                    self._LOG.info(f"New RMSE: {new_metric} Old RMSE: {optimal_metric}")
                    optimal_metric = new_metric
                    optimal_transformation = new_rotation
                    optimal_scale_factors += scale_plus
                    errors.append(new_metric)
                    break

                # SCALE TARGET CLOUD
                scale_neg = -self._delta * directions[:, axis]
                new_scale_factors, new_rotation, new_metric = self.compass_step(
                    source, target, optimal_scale_factors, scale_neg
                )

                if new_metric <= optimal_metric:
                    self._LOG.info(f"New RMSE: {new_metric} Old RMSE: {optimal_metric}")
                    optimal_metric = new_metric
                    optimal_transformation = new_rotation
                    optimal_scale_factors += scale_neg
                    errors.append(new_metric)
                    break
            # TODO (ADD try/except?)
            # UPDATE DELTA
            if new_metric > optimal_metric:
                self._delta = self._delta / 2
                self._LOG.info(f"No improvement, decreasing step size to {self._delta}")

        if refine_registration:
            target = target * optimal_scale_factors
            optimal_transformation, optimal_metric = self.refine_registration(
                source=source, target=target, initial_transform=optimal_transformation, icp_type=icp_type
            )
            errors.append(optimal_metric)

        if self._visualize_intermediate_steps:
            draw_registration_result(
                create_cloud(source),
                create_cloud(target * optimal_scale_factors),
                optimal_transformation,
            )

        self.transfromation = optimal_transformation
        self.scale_factors = optimal_scale_factors

        return optimal_transformation, optimal_metric, optimal_scale_factors, errors

    def refine_registration(
        self,
        source: np.ndarray,
        target: np.ndarray,
        initial_transform: np.ndarray,
        max_iteration: int = __REFINER_MAX_ITER__,
        distance_threshold: float = __REFINER_DISTANCE_THRESHOLD__,
        icp_type: str = "PointToPoint",
    ) -> (np.ndarray, float):
        """Refine the registration using ICP algorithm.
        :param source: Source point cloud
        :param target: Target point cloud
        :param initial_transform: Initial transformation matrix
        :param max_iteration: Maximum number of iterations for the ICP algorithm
        :param distance_threshold: Maximum distance between two correspondences
        :param icp_type: PointToPoint or PointToPlane ICP

        :return: (T_refined, result.inlier_rmse): (Refined transformation matrix, RMSE of the inliers)
        """
        if icp_type == "PointToPoint":
            icp_type = (
                o3d.pipelines.registration.TransformationEstimationPointToPoint()
            )  # for faster convergence
        if icp_type == "PointToPlane":
            icp_type = o3d.pipelines.registration.TransformationEstimationPointToPlane()
        # Define convergence criterion
        convergence_rule = o3d.pipelines.registration.ICPConvergenceCriteria(
            max_iteration=max_iteration
        )

        source_cloud = create_cloud(points=source)
        target_cloud = create_cloud(points=target)

        # Execute ICP
        result = o3d.pipelines.registration.registration_icp(
            source_cloud,
            target_cloud,
            distance_threshold,
            initial_transform,
            icp_type,
            convergence_rule,
        )

        T_refined = result.transformation

        return T_refined, result.inlier_rmse


    def transfrom(self, source: np.ndarray) -> np.ndarray:
        """Transform the source point cloud using the transformation matrix.
        :param source: Source point cloud
        :return: transformed_source: Transformed source point cloud to be aligned with the target
        """
        # Get the mean and distance of the source and target point clouds preprocessor
        for process_block in self._source_preprocessor.preprocessor_blocks:
            if issubclass(process_block.__class__, BaseScaler):
                source_mean = process_block.mean
                source_distance = process_block.scale
        for process_block in self._target_preprocessor.preprocessor_blocks:
            if issubclass(process_block.__class__, BaseScaler):
                target_mean = process_block.mean
                target_distance = process_block.scale

        # Normalize the source point cloud
        source = (source - source_mean) / source_distance

        # Rototranslate the source cloud as done in the align method
        source = np.dot(source, self.transfromation[:3, :3]) + self.transfromation[:3, 3]

        # Apply the inverse scale factors
        source = source / self.scale_factors

        # Denormalize the source point cloud with the target point cloud
        source = source * target_distance + target_mean

        return source

    def __repr__(self):
        return f"""{self.__class__.__name__}
            (source_preprocessor={self._source_preprocessor},
            target_preprocessor={self._target_preprocessor},
            optimizer={self._optimizer},
            attempts={self._attempts},
            deg={self._deg},
            mu={self._mu},
            std={self._std},
            delta={self._delta},
            max_iter={self._max_iter},
            eps={self._eps})
        """



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np


class IOptimizer(ABC):

    @abstractmethod
    def optimize(self, source: np.ndarray, target: np.ndarray, **kwargs) -> Tuple[np.ndarray, float]:
        """Method to run the inner block optimizer.

        Args:
            source (numpy array): source point-cloud array
            target (numpy array): target point-cloud array

        Returns:
            Tuple[np.ndarray, float]: A tuple with the Roto-translation matrix and inlier RMSE of the solution
        """
        pass



import numpy as np
from numpy import ndarray
from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.utils.logger_factory import LoggerFactory
import open3d as o3d

from or_pcd.utils.constants import (
    __NB_NEIGHBOURS__,
    __STD_RATIO__,
)


class SOR(IProcessBlock):
    """
    Statistical outlier removal.

    This class implements the statistical outlier removal (SOR) algorithm, which is used to filter Outliers from point cloud data.

    Args:
        nb_neighbours (int): Number of neighbours used by the SOR algorithm to compute mean and variance.
        std_ratio (float): SOR parameter. By increasing it, more points are kept as inliers.
    """

    def __init__(
        self, nb_neighbours: int = __NB_NEIGHBOURS__, std_ratio: float = __STD_RATIO__
    ) -> None:
        """
        Initialize the SOR object with the specified parameters.

        Args:
            nb_neighbours (int): Number of neighbours used by the SOR algorithm to compute mean and variance.
            std_ratio (float): SOR parameter. By increasing it, more points are kept as inliers.
        """

        super().__init__()

        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=True
        )

        self._nb_neighbours = nb_neighbours
        """
        Number of neighbours used by SOR algorithm (to compute mean and variance)
        """

        self._std_ratio = std_ratio
        """
        SOR parameter. By increasing it, more points are kept (as inliers)
        """

    def process(self, cloud: ndarray) -> ndarray:
        """
        Filters Outliers from the input point cloud using the SOR algorithm and returns only the inliers.
        Since the implementation provided by open3d is used, a temporary pointcluoud is created with
        the provided array

        Args:
            cloud (ndarray): Input point cloud data.

        Returns:
            ndarray: Filtered point cloud containing only the inliers.
        """

        # Create a temporary point cloud to use open3d method
        _temp = o3d.geometry.PointCloud()
        _temp.points = o3d.utility.Vector3dVector(np.copy(a=cloud))

        # Remove Outliers
        inliers, ind = _temp.remove_statistical_outlier(
            nb_neighbors=self._nb_neighbours, std_ratio=self._std_ratio
        )
        inliers = np.asarray(inliers.points)

        self._LOG.debug(
            msg=f"Cloud before SOR had {int(cloud.shape[0])} points. After SOR has {inliers.shape[0]} points!"
        )

        # Return only inliers
        return inliers

    @property
    def nb_neighbours(self) -> int:
        """
        Getter method for the number of neighbours parameter.

        Returns:
            int: Number of neighbours used by the SOR algorithm.
        """
        return self._nb_neighbours

    @property
    def std_ratio(self) -> float:
        """
        Getter method for the standard deviation ratio parameter.

        Returns:
            float: Standard deviation ratio used by the SOR algorithm.
        """
        return self._std_ratio

    def __repr__(self):
        return f"{self.__class__.__name__}(nb_neighbours={self._nb_neighbours}, std_ratio={self._std_ratio})"



from typing import List

import numpy as np
from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.Preprocessor.Scalers.BaseScaler import BaseScaler
from or_pcd.Preprocessor.Scalers import RadiusScaler

from or_pcd.utils.logger_factory import LoggerFactory


class Preprocessor:
    def __init__(self, preprocessor_blocks: List[IProcessBlock]):
        """Preprocessor class to preprocess the input cloud using the preprocessor blocks
        Scalers block should be always applied
        """

        self._LOG = LoggerFactory.get_logger(self.__class__.__name__, log_on_file=False)
        # Check if the Scalers class is among the preprocessor blocks
        if not any(issubclass(block.__class__, BaseScaler) for block in preprocessor_blocks):
            self._LOG.warning(
                """Scalers block is not present in the preprocessor blocks --> Adding Scalers block to the preprocessor blocks"""
            )
            preprocessor_blocks.insert(0, RadiusScaler())
        self.preprocessor_blocks = preprocessor_blocks

    def preprocess(self, cloud: np.ndarray) -> np.ndarray:
        """Preprocess the input cloud using the preprocessor blocks. sequentially"""
        # Preprocess data
        for block in self.preprocessor_blocks:
            cloud = block.process(cloud)
        return cloud



import numpy as np
import open3d as o3d

from or_pcd.utils.util_mcs_registration import generate_initialization_matrix

# Create a 5x3 array
source = np.random.randn(5, 3)

source_cloud = o3d.geometry.PointCloud()
source_cloud.points = o3d.utility.Vector3dVector(source)

T = generate_initialization_matrix(deg=1.57, std=0.1)

rotation_matrix = T[:3, :3]
translaton_vector = T[:3, 3]

# Transform with Open3D
source_cloud.transform(T)
trans_o3d = np.asarray(source_cloud.points)

# Transform with normal numpy
trans_normal = np.dot(source, rotation_matrix.T) + translaton_vector

print("Open3D: \n", trans_o3d, "\n\nNormal\n", trans_normal)



import numpy as np
from scipy.spatial.distance import cdist
from tqdm import tqdm

from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.utils.constants import __SAMPLE_SIZE__
from or_pcd.utils.logger_factory import LoggerFactory


class FarthestDownsampler(IProcessBlock):
    def __init__(self, sample_size: int = __SAMPLE_SIZE__):
        """
        :param sample_size: size of the down-sampled cloud
        """
        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )

        if sample_size <= 0:
            msg = f"sample size cannot be 0 or less. Provided: {sample_size}. Using default value: {__SAMPLE_SIZE__}"
            self._LOG.warning(msg)
            self._sample_size = __SAMPLE_SIZE__
        else:
            self._sample_size = sample_size

    def process(self, cloud: np.ndarray):
        """
        :param cloud: cloud to down-sample
        :return: cloud downsampled with farthest point algorithm refer to pointnet ++ paper for explanation
        link: https://arxiv.org/abs/1706.02413
        """
        # Create a list of indexes of the points that will be kept
        index_list = [0] * self._sample_size
        # The first point is chosen randomly
        index_list[0] = np.random.randint(low=0, high=cloud.shape[0])

        # Create a distance matrix with all points far away
        distances = np.ones((1, cloud.shape[0])) * 1e6

        # Iterate over the sample size and choose the farthest point
        for i in tqdm(range(self._sample_size - 1)):
            # Get the current point
            point = cloud[index_list[i]].reshape((1, cloud.shape[1]))

            # Calculate the distance between the current point and all the others
            point_distances = cdist(point, cloud)
            # Get the minimum distance between the current distances and the new ones
            distances = np.min((point_distances, distances), axis=0)

            # Get the farthest point among all distances
            fp_index = np.argmax(distances)
            index_list[i + 1] = fp_index

        return cloud[index_list]

    def __repr__(self):
        return f"{self.__class__.__name__}(sample_size={self._sample_size})"



from abc import ABC, abstractmethod

import numpy as np


class IProcessBlock(ABC):
    """Abstract class for a single inner block of preprocessing pipeline."""

    @abstractmethod
    def process(self, cloud: np.ndarray) -> np.ndarray:
        """Processing function of the input cloud array."""
        pass



import open3d as o3d
import pkg_resources
import numpy as np

def load_sample_cloud(name: str = "ArmadilloBack_180") -> np.ndarray:
    """
    Load a sample cloud from the or_pcd/data directory

    Args:
        name (str): Name of the cloud to load

    Returns:
        o3d.geometry.PointCloud: The loaded point cloud
    """
    path = name + ".ply"
    data_path = pkg_resources.resource_stream(__name__, path)
    cloud: o3d.geometry.PointCloud = o3d.io.read_point_cloud(data_path.name)
    # Return the array of the cloud
    return np.asarray(cloud.points)


from or_pcd.Preprocessor.Downsamplers.farthestDownsampler import FarthestDownsampler
from or_pcd.Preprocessor.Downsamplers.randomDownsampler import RandomDownsampler
from or_pcd.Preprocessor.Downsamplers.voxelDownsampler import VoxelDownsampler




import copy
from typing import Tuple

import numpy as np
import open3d as o3d

from or_pcd.Optimizer.iOptimizer import IOptimizer
from or_pcd.utils.constants import (
    __DIVISION_FACTOR__,
    __TUPLE_SCALE__,
    __MAXIMUM_CORRESPONDENCE_DISTANCE__,
    __ITERATION_NUMBER__,
    __DECREASE_MU__,
    __FPFH_RADIUS__,
    __FPFH_KNN__,
    __NORMAL_ESTIMATE_RADIUS__,
    __NORMAL_ESTIMATE_KNN__,
)
from or_pcd.utils.logger_factory import LoggerFactory


class FastGlobalOptimizer(IOptimizer):
    def __init__(
        self,
        division_factor: float = __DIVISION_FACTOR__,
        tuple_scale: float = __TUPLE_SCALE__,
        maximum_correspondence_distance: float = __MAXIMUM_CORRESPONDENCE_DISTANCE__,
        iteration_number: int = __ITERATION_NUMBER__,
        decrease_mu: bool = __DECREASE_MU__,
        normal_estimate_radius: float = __NORMAL_ESTIMATE_RADIUS__,
        normal_estimate_knn: int = __NORMAL_ESTIMATE_KNN__,
        fpfh_radius: float = __FPFH_RADIUS__,
        fpfh_knn: int = __FPFH_KNN__,
    ):
        """
        Refer to this link: https://vladlen.info/papers/fast-global-registration.pdf for parameters explanation
        :param division_factor:
        :param tuple_scale:
        :param maximum_correspondence_distance:
        :param iteration_number:
        :param decrease_mu:
        :param fpfh_radius:
        :param fpfh_knn:
        :param fpfh_radius_features:
        """
        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )
        if division_factor <= 0:
            msg = f"division factor cannot be 0 or less. Provided: {division_factor}"
            self._LOG.error(msg)
            self._division_factor = __DIVISION_FACTOR__
        else:
            self._division_factor = division_factor

        if tuple_scale <= 0:
            msg = f"tuple scale cannot be 0 or less. Provided: {tuple_scale}"
            self._LOG.error(msg)
            self._tuple_scale = __TUPLE_SCALE__
        else:
            self._tuple_scale = tuple_scale

        if maximum_correspondence_distance <= 0:
            msg = f"maximum correspondence distance cannot be 0 or less. Provided: {maximum_correspondence_distance}"
            self._LOG.error(msg)
            self._maximum_correspondence_distance = __MAXIMUM_CORRESPONDENCE_DISTANCE__
        else:
            self._maximum_correspondence_distance = maximum_correspondence_distance

        if iteration_number <= 0:
            msg = f"iteration number cannot be 0 or less. Provided: {iteration_number}"
            self._LOG.error(msg)
            self._iteration_number = __ITERATION_NUMBER__
        else:
            self._iteration_number = iteration_number

        self._decrease_mu = decrease_mu

        if fpfh_radius <= 0:
            msg = f"fpfh radius cannot be 0 or less. Provided: {fpfh_radius}"
            self._LOG.error(msg)
            self._fpfh_radius = __FPFH_RADIUS__
        else:
            self._fpfh_radius = fpfh_radius

        if fpfh_knn <= 0:
            msg = f"fpfh knn cannot be 0 or less. Provided: {fpfh_knn}"
            self._LOG.error(msg)
            self._fpfh_knn = __FPFH_KNN__
        else:
            self._fpfh_knn = fpfh_knn

        if normal_estimate_radius <= 0:
            msg = f"normal estimate radius cannot be 0 or less. Provided: {normal_estimate_radius}"
            self._LOG.error(msg)
            self._normal_estimate_radius = __NORMAL_ESTIMATE_RADIUS__
        else:
            self._normal_estimate_radius = normal_estimate_radius

        if normal_estimate_knn <= 0:
            msg = f"normal estimate knn cannot be 0 or less. Provided: {normal_estimate_knn}"
            self._LOG.error(msg)
            self._normal_estimate_knn = __NORMAL_ESTIMATE_KNN__
        else:
            self._normal_estimate_knn = normal_estimate_knn

        self._LOG.debug(msg=f"initialized optimizer: {self}")

    def get_fpfh_features(
        self,
        source_point_cloud: o3d.geometry.PointCloud,
        target_point_cloud: o3d.geometry.PointCloud,
    ) -> Tuple[o3d.pipelines.registration.Feature, o3d.pipelines.registration.Feature]:
        source_copy = copy.deepcopy(source_point_cloud)
        target_copy = copy.deepcopy(target_point_cloud)

        # Estimate Normals
        source_copy.estimate_normals(
            o3d.geometry.KDTreeSearchParamHybrid(
                radius=self._normal_estimate_radius, max_nn=self._normal_estimate_knn
            )
        )
        target_copy.estimate_normals(
            o3d.geometry.KDTreeSearchParamHybrid(
                radius=self._normal_estimate_radius, max_nn=self._normal_estimate_knn
            )
        )

        # Estimate FPFH Features
        source_fpfh_features = o3d.pipelines.registration.compute_fpfh_feature(
            source_copy,
            o3d.geometry.KDTreeSearchParamHybrid(
                radius=self._fpfh_radius, max_nn=self._fpfh_knn
            ),
        )

        target_fpfh_features = o3d.pipelines.registration.compute_fpfh_feature(
            source_copy,
            o3d.geometry.KDTreeSearchParamHybrid(
                radius=self._fpfh_radius, max_nn=self._fpfh_knn
            ),
        )

        return source_fpfh_features, target_fpfh_features

    def optimize(
        self, source: np.ndarray, target: np.ndarray, **kwargs
    ) -> Tuple[np.ndarray, float]:
        source_point_cloud = o3d.geometry.PointCloud()
        target_point_cloud = o3d.geometry.PointCloud()
        source_point_cloud.points = o3d.utility.Vector3dVector(source)
        target_point_cloud.points = o3d.utility.Vector3dVector(target)

        source_fpfh_features, target_fpfh_features = self.get_fpfh_features(
            source_point_cloud, target_point_cloud
        )

        fast_global_option = o3d.pipelines.registration.FastGlobalRegistrationOption(
            division_factor=self._division_factor,
            tuple_scale=self._tuple_scale,
            maximum_correspondence_distance=self._maximum_correspondence_distance,
            iteration_number=self._iteration_number,
            decrease_mu=self._decrease_mu,
        )

        optimization_result: o3d.pipelines.registration.RegistrationResult = (
            o3d.pipelines.registration.registration_fgr_based_on_feature_matching(
                source_point_cloud,
                target_point_cloud,
                source_fpfh_features,
                target_fpfh_features,
                fast_global_option,
            )
        )

        roto_translation = np.copy(optimization_result.transformation)
        transposed_rotation = roto_translation[:3, :3].T
        roto_translation[:3, :3] = transposed_rotation

        # If the rmse is 0 raise an error as parameters are not correct
        if len(optimization_result.correspondence_set) == 0:
            msg = """No correspondences detected from the optimizer. Parameters are not well set. 
        Probably due to: 
        -   maximum_correspondence_distance set too low
        -   tuple_scale set too high
        """
            self._LOG.error(msg)
            raise Warning(msg)

        return roto_translation, optimization_result.inlier_rmse

    def __repr__(self):
        return f"""{self.__class__.__name__}
            (division_factor={self._division_factor},
            tuple_scale={self._tuple_scale},
            maximum_correspondence_distance={self._maximum_correspondence_distance},
            iteration_number={self._iteration_number},
            decrease_mu={self._decrease_mu},
            normal_estimate_radius={self._normal_estimate_radius},
            normal_estimate_knn={self._normal_estimate_knn},
            fpfh_radius={self._fpfh_radius},
            fpfh_knn={self._fpfh_knn})"
        """



import os
import time
from datetime import datetime
from pathlib import Path
import math

# TIMESTAMP (COMPUTED ONCE FOR LOGGERS INITIALIZATION)
__TIME_NOW__ = time.time()

# TIMESTAMP EXPRESSED IN HUMAN READABLE FORMAT
__DATETIME_NOW__ = formatted_time = datetime.fromtimestamp(__TIME_NOW__).strftime("%Y_%m_%d_%H_%M_%S")

# PROJECT ROOT
__ROOT__ = str(Path(os.getcwd()).parent)

# LOG DEFAULT FOLDER
__LOG_FOLDER__ = os.path.join(__ROOT__, "logs", __DATETIME_NOW__)

# LOG DEFAULT FORMAT
__DEFAULT_LOG_FORMAT__ = "%(asctime)s | %(name)s | %(levelname)s : %(message)s"

# ------------------------------------------------------------------------------------------------------- #
# OUTLIER REMOVAL
# ------------------------------------------------------------------------------------------------------- #
# SOR
__NB_NEIGHBOURS__ = 64
"""
Number of neighbours used by SOR algorithm (to compute mean and variance)
"""
__STD_RATIO__ = 2
"""
SOR parameter. By increasing it, more points are kept (as inliers)
"""
# ------------------------------------------------------------------------------------------------------- #

# DOWN SAMPLERS
# RANDOM DOWN SAMPLER
__RANDOM_SAMPLE_SIZE__ = 15000

# FARTHEST DOWN SAMPLER
__FARTHEST_SAMPLE_SIZE__ = 10000

# VOXEL DOWN SAMPLER
__VOXEL_SAMPLE_SIZE__ = 2000
__BASE_VOXEL_SIZE__ = 0.01
__MIN_VOXEL_SIZE__ = 0.0001
__DELTA__ = 0.01
__EPS__ = 0.0005

# DOWNSAMPLER DEFAULT SAMPLE SIZE
__SAMPLE_SIZE__ = 4096

# FAST GLOBAL OPTIMIZER
__DIVISION_FACTOR__ = 1.4
__TUPLE_SCALE__ = 0.9
__ITERATION_NUMBER__ = 100
__DECREASE_MU__ = True
__NORMAL_ESTIMATE_RADIUS__ = 0.1
__NORMAL_ESTIMATE_KNN__ = 20
__FPFH_RADIUS__ = 0.1
__FPFH_KNN__ = 20

# GENERALIZED ICP
__MAX_ITERATIONS__ = 100

# OPTIMIZER GENERAL PARAMS
__MAXIMUM_CORRESPONDENCE_DISTANCE__ = 0.5


# ALIGNER
__MULTISTART_ATTEMPTS__ = 30
__ALIGNER_DEG__ = math.pi / 2
__ALIGNER_MU__ = 0.0
__ALIGNER_STD__ = 0.1
__ALIGNER_MAX_ITER__ = 100
__ALIGNER_DELTA__ = 0.2
__ALIGNER_EPS__ = 0.05

# REFINER
__REFINER_MAX_ITER__ = 200
__REFINER_DISTANCE_THRESHOLD__ = 0.5



import numpy as np
import open3d as o3d

from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.utils.constants import (
    __BASE_VOXEL_SIZE__,
    __MIN_VOXEL_SIZE__,
    __DELTA__,
    __EPS__,
)
from or_pcd.utils.logger_factory import LoggerFactory


class VoxelDownsampler(IProcessBlock):
    """
    Reduces point cloud dimension by creating a voxel of the space. To select the appropriate voxel size, it uses
    a pattern search algorithm that, given the desired number of points, finds the appropriate voxel size
    """

    def __init__(
        self,
        target_points: int,
        base_voxel_size: float = __BASE_VOXEL_SIZE__,
        min_voxel_size: float = __MIN_VOXEL_SIZE__,
        delta: float = __DELTA__,
        eps: float = __EPS__,
    ):
        """
        :param target_points: desired number of points after down sampling
        :param base_voxel_size:
        :param min_voxel_size: minimum voxel size that can be used. It must be greater than 0
        :param delta:
        :param eps:
        """

        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )

        if target_points <= 0:
            msg = f"target points cannot be 0 or less. Provided: {target_points}"
            self._LOG.error(msg)
            raise ValueError(msg)

        self._target_points = target_points

        if min_voxel_size <= 0:
            msg = f"min voxel size cannot be 0 or less. Provided: {min_voxel_size}. Using default value: {__MIN_VOXEL_SIZE__}"
            self._LOG.warning(msg)
            self._min_voxel_size = __MIN_VOXEL_SIZE__
        else:
            self._min_voxel_size = min_voxel_size

        if base_voxel_size <= 0:
            msg = f"base voxel size cannot be 0 or less. Provided: {base_voxel_size}. Using default value: {__BASE_VOXEL_SIZE__}"
            self._LOG.warning(msg)
            self._base_voxel_size = __BASE_VOXEL_SIZE__
        else:
            self._base_voxel_size = base_voxel_size

        if delta <= 0:
            msg = f"delta cannot be 0 or less. Provided: {delta}. Using default value: {__DELTA__}"
            self._LOG.warning(msg)
            self._delta = __DELTA__
        else:
            self._delta = delta

        if eps <= 0:
            msg = f"eps cannot be 0 or less. Provided: {eps}. Using default value: {__EPS__}"
            self._LOG.warning(msg)
            self._eps = __EPS__
        else:
            self._eps = eps

        self._LOG.debug(msg=f"initialized downsampler: {self}")

    def compass_step(
        self, delta: float, cloud: o3d.geometry.PointCloud, current_voxel_size: float
    ):
        new_voxel_size = current_voxel_size + delta
        if new_voxel_size <= self._min_voxel_size:
            new_voxel_size = self._min_voxel_size
        cloud_ds = cloud.voxel_down_sample(new_voxel_size)
        n_points = np.asarray(cloud_ds.points).shape[0]
        metric = np.abs(self._target_points - n_points)
        return cloud_ds, new_voxel_size, metric

    def process(self, cloud: np.ndarray):
        """Get the voxel size of the cloud needed for target sample size"""
        current_voxel_size = self._base_voxel_size
        source_point_cloud = o3d.geometry.PointCloud()
        source_point_cloud.points = o3d.utility.Vector3dVector(cloud)

        # Downsample first run
        best_cloud = source_point_cloud.voxel_down_sample(current_voxel_size)
        # Get number of points
        n_points = np.asarray(best_cloud.points).shape[0]
        # Compute Metric
        metric = np.abs(self._target_points - n_points)

        while self._delta >= self._eps:
            obtained_cloud, new_voxel_size, obtained_metric = self.compass_step(
                self._delta, source_point_cloud, current_voxel_size
            )

            if obtained_metric < metric:
                current_voxel_size = new_voxel_size
                metric = obtained_metric
                best_cloud = obtained_cloud
                continue

            obtained_cloud, new_voxel_size, obtained_metric = self.compass_step(
                -self._delta, source_point_cloud, current_voxel_size
            )

            if obtained_metric < metric:
                current_voxel_size = new_voxel_size
                metric = obtained_metric
                best_cloud = obtained_cloud
                continue

            self._delta = self._delta / 2

        # Log the voxel size and the number of points
        self._LOG.debug(msg=f"Voxel size: {current_voxel_size}, Metric: {metric}")
        return np.asarray(best_cloud.points)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"target_points={self._target_points}, "
            f"base_voxel_size={self._base_voxel_size}, "
            f"min_voxel_size={self._min_voxel_size}, "
            f"delta={self._delta}, "
            f"eps={self._eps})"
        )



import numpy as np
import open3d as o3d


def create_cloud(points: np.ndarray) -> o3d.geometry.PointCloud:
    """
    Shortcut to create an open3d point cloud from a numpy array
    """

    cloud = o3d.geometry.PointCloud()
    cloud.points = o3d.utility.Vector3dVector(points)
    return cloud



import numpy as np

from or_pcd.Preprocessor.iProcessBlock import IProcessBlock
from or_pcd.utils.constants import __SAMPLE_SIZE__
from or_pcd.utils.logger_factory import LoggerFactory


class RandomDownsampler(IProcessBlock):
    def __init__(self, sample_size: int = __SAMPLE_SIZE__, replace: bool = False):
        """
        :param sample_size: the number of points to keep from the cloud
        :param replace: choose to sample with or without replacement
        """
        self._LOG = LoggerFactory.get_logger(
            log_name=self.__class__.__name__, log_on_file=False
        )

        if sample_size <= 0:
            msg = f"sample size cannot be 0 or less. Provided: {sample_size}. Using default value: {__SAMPLE_SIZE__}"
            self._LOG.warning(msg)
            self._sample_size = __SAMPLE_SIZE__
        else:
            self._sample_size = sample_size

        self._replace = replace

        self._LOG.debug(f"Initialized {self.__class__.__name__} with {self}")

    def process(self, cloud: np.ndarray) -> np.ndarray:
        """
        :param cloud: cloud to down-sample
        :return: randomly down-sampled cloud
        """
        # Downsample data
        return cloud[
            np.random.choice(cloud.shape[0], self._sample_size, replace=self._replace)
        ]

    def __repr__(self):
        return f"{self.__class__.__name__}(sample_size={self._sample_size}, replace={self._replace})"



import copy
from typing import Iterable, Tuple

import numpy as np
import open3d as o3d


class Visualizer:
    # TODO create this class
    def __init__(self, data):
        self.data = data

    def visualize(self):
        # Visualize the data
        pass


def visualize_point_clouds(
    clouds_points: Iterable[np.ndarray],
    colors_rgb: Iterable[Tuple[float, float, float]],
) -> None:
    """
    TODO: Generate docstring with copilot
    """

    clouds = []
    for color_rgb, cloud_points in zip(colors_rgb, clouds_points):
        cloud_copy = cloud_points.copy()
        cloud = o3d.geometry.PointCloud()
        cloud.points = o3d.utility.Vector3dVector(cloud_copy)
        cloud.paint_uniform_color(color=color_rgb)
        clouds.append(cloud)

    o3d.visualization.draw_geometries(clouds)


def draw_registration_result(
    source: o3d.geometry.PointCloud, target: o3d.geometry.PointCloud, transformation
):
    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)
    source_temp.paint_uniform_color([0.0, 0.0, 1.0])
    target_temp.paint_uniform_color([1.0, 0.0, 0.0])

    # Rotate the sub 3x3 matrix of transformation matrix
    transform = np.copy(transformation)
    transform[:3, :3] = transformation[:3, :3].T
    source_temp.transform(transform)

    o3d.visualization.draw_geometries([source_temp, target_temp])


def np_draw_registration_result(
    source: np.ndarray, target: np.ndarray, transformation: np.ndarray
):
    source_transformerd = np.dot(source, transformation[:3, :3]) + transformation[:3, 3]

    source_cloud = o3d.geometry.PointCloud()
    target_cloud = o3d.geometry.PointCloud()
    source_cloud.points = o3d.utility.Vector3dVector(source_transformerd)
    target_cloud.points = o3d.utility.Vector3dVector(target)

    source_cloud.paint_uniform_color([0.0, 0.0, 1.0])
    target_cloud.paint_uniform_color([1.0, 0.0, 0.0])

    o3d.visualization.draw_geometries([source_cloud, target_cloud])



from setuptools import find_packages, setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name = "or-pcd",
    version = "0.0.1-BETA.7",
    description = "A python package to perform pointcloud registration with scale differences using a two-block optimization approach",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dpolimeni/Multi-scale-pointcloud-registration",
    author="Diego Polimeni and Alessandro Pannone",
    author_email="a.pannone1798@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy==1.26.4",
        "scikit-learn==1.4.1.post1",
        "open3d==0.18.0"
    ],
    package_data={"": ["data/*.conf", "data/*.ply"]},
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.11",
) 


import inspect
import logging
import os
import sys
from logging import Logger, FileHandler, Formatter, StreamHandler

from or_pcd.utils.constants import __LOG_FOLDER__, __DEFAULT_LOG_FORMAT__


class LoggerFactory:
    """
    Use this class to generate custom loggers!
    """

    def __init__(self):
        pass

    @staticmethod
    def get_logger(
        log_name: str = None,
        log_on_file: bool = False,
        file_name: str = None,
        folder_name: str = None,
        log_file_format: str = None,
        log_console_format: str = None,
        file_handler_level: int = logging.DEBUG,
        console_handler_level: int = logging.DEBUG,
    ) -> Logger:
        """
        Creates customized logger

        :param log_name: name of the logger. If not provided, the file from which the method is called will be used
        :param log_on_file: set this to True if you wish to log on file
        :param file_name: name of the log file. If not provided, the name of the logger will be used
        :param folder_name: name of the folder in which log file is placed. If not provided, __LOG_FOLDER__ is used
        :param log_file_format: log format for file. If None, __DEFAULT_LOG_FORMAT__ will be used
        :param log_console_format: log format for console. If None, __DEFAULT_LOG_FORMAT__ will be used
        :param file_handler_level: only messages above this level will be logged on file
        :param console_handler_level: only messages above this level will be logged on console
        """

        # GET LOGGER NAME (USE CALLER FILE IF NOT PROVIDED)
        _log_name = (
            log_name
            if log_name is not None
            else os.path.splitext(os.path.basename(inspect.stack()[1].filename))[0]
        )

        # INITIALIZE LOGGER
        logger = logging.getLogger(name=_log_name)
        logger.setLevel(level=logging.DEBUG)

        if log_on_file:

            # GET LOG FOLDER NAME (USE DEFAULT IF NOT PROVIDED)
            _folder_name = folder_name if folder_name is not None else __LOG_FOLDER__

            # CREATE DIRECTORY (IF NOT ALREADY AVAILABLE)
            if not os.path.isdir(s=_folder_name):
                os.makedirs(name=_folder_name, exist_ok=True)

            # GET LOG FILE NAME (USE CALLER FILE IF NOT PROVIDED)
            _file_name = file_name if file_name is not None else _log_name

            # CREATE FILE HANDLER
            file_handler = FileHandler(filename=os.path.join(_folder_name, _file_name))

            # SET LEVEL TO HANDLER
            file_handler.setLevel(level=file_handler_level)

            # CREATE AND SET FORMATTER FOR FILE HANDLER
            _log_file_format = (
                log_file_format
                if log_file_format is not None
                else __DEFAULT_LOG_FORMAT__
            )
            file_handler.setFormatter(fmt=Formatter(fmt=_log_file_format))

            # ADD FILE HANDLER TO LOGGER
            logger.addHandler(hdlr=file_handler)

        # INITIALIZE CONSOLE HANDLER
        console_handler = StreamHandler(stream=sys.stdout)
        console_handler.flush = sys.stdout.flush

        # SET LEVEL TO HANDLER
        console_handler.setLevel(level=console_handler_level)

        # CREATE AND SET FORMATTER FOR STREAM HANDLER
        _log_console_format = (
            log_console_format
            if log_console_format is not None
            else __DEFAULT_LOG_FORMAT__
        )
        console_handler.setFormatter(fmt=Formatter(fmt=_log_console_format))

        # ADD CONSOLE HANDLER TO LOGGER
        logger.addHandler(hdlr=console_handler)

        return logger
